/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2010 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "SyncQueueManager.h"

USE_NAMESPACE;

const int SyncQueueManager::sem_key_id          = 's';
const int SyncQueueManager::queued_shm_key_id   = 'm';
const int SyncQueueManager::current_shm_key_id  = 'M';
const int SyncQueueManager::sem_perms           = 0600;
const int SyncQueueManager::shm_perms           = 0600;
const int SyncQueueManager::shm_size            = sizeof(int);

SyncQueueManager::SyncQueueManager() : sync_info(0), queued_shm_val(NULL), current_shm_val(NULL), 
    currentSyncInfoSet(false), queuedSyncInfoSet(false), syncLockSet(false), syncLockRelinquished(false)
{
    key_path = PlatformAdapter::getConfigFolder();
    sem_key = ftok(key_path, sem_key_id);
    queued_shm_key  = ftok(key_path, queued_shm_key_id);
    current_shm_key = ftok(key_path, current_shm_key_id);
}


SyncQueueManager::SyncQueueManager(unsigned int info) : sync_info(info), queued_shm_val(NULL), current_shm_val(NULL), 
    currentSyncInfoSet(false), queuedSyncInfoSet(false), syncLockSet(false), syncLockRelinquished(false)
{
    key_path = PlatformAdapter::getConfigFolder();
    sem_key = ftok(key_path, sem_key_id);
    queued_shm_key  = ftok(key_path, queued_shm_key_id);
    current_shm_key = ftok(key_path, current_shm_key_id);
}

SyncQueueManager::~SyncQueueManager()
{
    cleanUpSyncQueue();
}

int SyncQueueManager::lockSyncQueue()
{
    if ((sem_id = semget(sem_key, 1, IPC_CREAT | IPC_EXCL | sem_perms)) < 0)  {
        if (errno == EEXIST) {
            int queueLockValid = 0;
            
            if ((sem_id = semget(sem_key, 1, 0)) < 0) {
                return -1;
            }

            syncLockSet = true;
            
            // check if semaphore is still valid
            if ((queueLockValid = checkLockStatus()) < 0) {
                return -1;
            }

            if (queueLockValid) {
                return 1;
            } 
        } else {
            return -1;
        }
    } else {
        sb.sem_num = 0;
        sb.sem_op  = 1; 
        sb.sem_flg = 0;
            
        if (semop(sem_id, &sb, 1) < 0) {
            return -1;        
        }
        syncLockSet = true;
    }
    
    return 0;
}

int SyncQueueManager::unlockSyncQueue()
{
    sb.sem_num = 0;
    sb.sem_op = 1;  
  
    if (semop(sem_id, &sb, 1) < 0) {
        LOG.error("error removing sync queue lock: %s", strerror(errno));

        return errno;
    }

    return 0;
}

int SyncQueueManager::checkLockStatus()
{
    pid_t sync_proc = 0;
    int queue_len = 0;

    // check if there is any process waiting for this semaphore
    if ((queue_len = semctl(sem_id, 0, GETNCNT)) < 0) {
        LOG.error("error in checking sync lock status: %s", strerror(errno));
        
        return -1;
    }
    
    if (queue_len) {
        return 1;
    }
    
    // check if last process which made a semop call is still running
    if ((sync_proc = semctl(sem_id, 0, GETPID)) < 0) {
        LOG.error("error in checking sync lock status: %s", strerror(errno));
        
        return -1;
    }

    if (sync_proc != getpid()) {
        if (checkProcessRunningByPid(sync_proc)) {
            return 1;
        }
    }
    
    return 0;
}

int SyncQueueManager::setCurrentSyncType()
{
    int *shm_val = 0;
    
    if ((queued_shm_id = shmget(queued_shm_key, shm_size, shm_perms | IPC_CREAT | IPC_EXCL)) < 0) {
        if (errno == EEXIST) {
            if ((queued_shm_id = shmget(queued_shm_key, shm_size, shm_perms | IPC_R | IPC_W)) < 0) {
                LOG.error("can't open shared memory: %s", strerror(errno));
                
                return errno;
            }
        } else {
            LOG.error("can't create shared memory: %s", strerror(errno));
        
            return errno;
        }
    }

    if ((queued_shm_val = (int *)shmat(queued_shm_id, NULL, 0)) == (int *)-1) {
        LOG.error("error attaching to shared memory: %s", strerror(errno));

        return errno;
    }
    
    // check value to be set in shared memory: 
    // at least one of the sync source must be set 
    if (((sync_info & SYNC_SOURCE_CONFIG) == 0) &&
		((sync_info & SYNC_SOURCE_CONTACTS) == 0) &&
		((sync_info & SYNC_SOURCE_BRIEFCASE) == 0)) {
		
        shmdt(queued_shm_val);
        LOG.error("invalid sync information type: %d", sync_info);

        return EINVAL;        
    }

    *queued_shm_val |= sync_info;
    
    // set flags and remain attached to segment
    queuedSyncInfoSet = true;
    
    sb.sem_num = 0;
    sb.sem_op = -1;  
    sb.sem_flg = SEM_UNDO;
    
    if (semop(sem_id, &sb, 1) < 0) {
        LOG.error("error activating sync queue lock: %s", strerror(errno));
        
        return errno;
    }

    // check if sync process interrupted by signal
    if (syncLockRelinquished) {
        LOG.info("sync queue lock relinquished");
        return 0; 
    }
    
    // semaphore unlocked: write current sync info and start sync
    if ((current_shm_id = shmget(current_shm_key, shm_size, shm_perms | IPC_CREAT | IPC_EXCL)) < 0) {
        if (errno == EEXIST) {
            struct shmid_ds shmds;
            
            if ((current_shm_id = shmget(current_shm_key, shm_size, shm_perms | IPC_R | IPC_W)) < 0) {
                LOG.error("can't open shared memory: %s", strerror(errno));
                
                return errno;
            }
            
            memset(&shmds, 0, sizeof(struct shmid_ds));
            
            if (shmctl(current_shm_id, IPC_STAT, &shmds) < 0) {
                LOG.error("error stating shared memory: %s", strerror(errno));
                
                return errno;
            }
            
            // no process should be attached to this memory segment now
            if (shmds.shm_nattch != 0) {
                LOG.error("current sync info shared memory has still some process attached");
                return EEXIST;
            }
            
        } else {
            LOG.error("can't create shared memory: %s", strerror(errno));
            
            return errno;
        }
    }
    
    if ((current_shm_val = (int *)shmat(current_shm_id, NULL, 0)) == (int *)-1) {
        LOG.error("error attaching to shared memory: %s", strerror(errno));
        
        return errno;
    }
    
    // set flag and remain attached to segment
    *current_shm_val = sync_info;
    currentSyncInfoSet = true;
    
    return 0;           
}

int SyncQueueManager::getCurrentSyncType()
{
    int shm_id = 0;
    int *sync_val = 0;
    int ret = 0;
    struct shmid_ds shmds;
    key_t shm_key = ftok(PlatformAdapter::getConfigFolder(), current_shm_key_id);
    
    if ((shm_id = shmget(shm_key, shm_size, shm_perms | IPC_R)) < 0) {
        // no sync is running
        return SYNC_SOURCE_NONE;
    }
 
    memset(&shmds, 0, sizeof(struct shmid_ds));
    
    if (shmctl(shm_id, IPC_STAT, &shmds) < 0) {
        LOG.error("error stating shared memory: %s", strerror(errno));
        
        return -1;
    }
    
    // no process is attached to shared memory: no sync is running
    // and this info was probably left by a process died badly
    if (shmds.shm_nattch == 0) {
        return SYNC_SOURCE_NONE;
    }
        
    if ((sync_val  = (int *)shmat(shm_id, NULL, 0)) == (int *)-1) {
        LOG.error("error attaching to shared memory: %s", strerror(errno));
        
        return -1;
    }
    
    ret = *sync_val;
    
    shmdt(sync_val);
    
    return ret;
}

bool SyncQueueManager::isSyncRunning()
{
    int syncType = SYNC_SOURCE_NONE;
    
    if ((syncType = getCurrentSyncType()) < 0) {
        LOG.error("error getting sync queue status");
        
        return false;
    }
    
    return (syncType != SYNC_SOURCE_NONE);
}

bool SyncQueueManager::isContactSyncRunning()
{
    int syncType = SYNC_SOURCE_NONE;
    
    if ((syncType = getCurrentSyncType()) < 0) {
        LOG.error("error getting sync queue status");
        
        return false;
    }
    
    return ((syncType & SYNC_SOURCE_CONTACTS) &&
            ((syncType & SYNC_SOURCE_RESET) == 0));
}

bool SyncQueueManager::isContactResetRunning()
{
    int syncType = SYNC_SOURCE_NONE;
    
    if ((syncType = getCurrentSyncType()) < 0) {
        LOG.error("error getting sync queue status");
        
        return false;
    }
    
    return ((syncType & SYNC_SOURCE_CONTACTS) &&
            (syncType & SYNC_SOURCE_RESET));
}

int SyncQueueManager::checkSyncQueue()
{
    int *sync_val = 0;
    int ret = 0;
    struct shmid_ds shmds;

    if ((queued_shm_id = shmget(queued_shm_key, shm_size, shm_perms | IPC_R)) < 0) {
        LOG.error("can't get shared memory access: %s", strerror(errno));
    
        return -1;
    }

    memset(&shmds, 0, sizeof(struct shmid_ds));
    
    if (shmctl(queued_shm_id, IPC_STAT, &shmds) < 0) {
        LOG.error("error stating shared memory: %s", strerror(errno));
        
        return -1;
    }
    
    // no process is attached to shared memory: no sync is running
    // and this info was probably left by a process died badly
    if (shmds.shm_nattch == 0) {
        return SYNC_SOURCE_NONE;
    }
    
    if ((sync_val  = (int *)shmat(queued_shm_id, NULL, 0)) == (int *)-1) {
        LOG.error("error attaching to shared memory: %s", strerror(errno));

        return -1;
    }

    // check if our flags has at least on of the sources which are already set
    if (*sync_val & sync_info) 
        ret = 1;

    shmdt(sync_val);

    return ret;
}

int SyncQueueManager::cleanUpSyncInfo()
{
    struct shmid_ds shmds;
      
    if (!queuedSyncInfoSet && !currentSyncInfoSet) {
        return 0;
    }
    
    // reset our queue sync flags
    *queued_shm_val ^= sync_info;
    shmdt(queued_shm_val);

    memset(&shmds, 0, sizeof(struct shmid_ds));
    
    if (shmctl(queued_shm_id, IPC_STAT, &shmds) < 0) {
        LOG.error("error stating shared memory: %s", strerror(errno));
        
        return errno;
    }
    
    if (shmds.shm_nattch == 0) {
        if (shmctl(queued_shm_id, IPC_RMID, 0)) {
            LOG.error("error removing shared memory: %s", strerror(errno));
            return errno;
        }
    }
    
    if (currentSyncInfoSet) {
        // reset current sync flags and remove memory segment
        *current_shm_val = 0;;
        shmdt(current_shm_val);

        if (shmctl(current_shm_id, IPC_RMID, 0)) {
            LOG.error("error removing shared memory: %s", strerror(errno));
            return errno;
        }
    }
    
    return 0;
}

int SyncQueueManager::cleanUpSyncQueue()
{
    int ret = 0;
    
    // remove shared memory info
    ret = cleanUpSyncInfo();
    
    if (syncLockSet) {
        // remove lock
        unlockSyncQueue();
     
        // check if semaphore is still in use
        if (checkLockStatus() == 0) {
            // remove semaphore if no one is using 
            // or waiting for it
            if (semctl(sem_id, 0, IPC_RMID) < 0) {
                LOG.error("error trying to remove semaphore: %s", strerror(errno));
                ret = errno;
            }
        }
    }
    
    return ret;
}

int SyncQueueManager::resetSyncQueue()
{
    syncLockRelinquished = true;

    return cleanUpSyncQueue();
}