/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "customization.h"
#include "constants.h"
#include "utils.h"

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <err.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


int execProcess(char * const args[], bool waitForChild)
{
    int res = 0;
    pid_t pid;
    int options = (waitForChild ? 0 : WNOHANG);

    switch (pid = fork()) {
        case -1:
            //NSLog(@"can't fork: %s\n", strerror(errno));
            return errno;
            
        case 0: {
            execv(args[0], args);
            
            
            exit(EXIT_FAILURE);
        }
            
        default: {
            int status = 0;
            
            if (waitpid(pid, &status, options) < 0) {
               // NSLog(@"error in waitpid: %s\n", strerror(errno));
                res = errno;
            } else {
                if (!WIFEXITED(status)) {
                   // NSLog(@"error executing \"%s\" command\n", args[0]);
                    res = ECHILD;
                } else {
                    if (WEXITSTATUS(status) != 0) {
                      //  NSLog(@"process \"%s\" exited with errors\n", args[0]);
                        res = ECHILD;
                    }
                }
            }
            
            break;
        }
    }
    
    return res;
}

int checkProcessRunning(const char* processName, int excludeSelf)
{
    struct kinfo_proc *proc_list = NULL;
    size_t proc_num = 0;
    int ret = 0;
    pid_t pid = 0;

    if (excludeSelf) {
        pid = getpid();
    }

    if ((processName != NULL) && (strlen(processName))) {
        if (getProcessList(&proc_list, &proc_num)) {
            ret = -1;
        } else {
            int i = 0;
            for (; i < proc_num; i++) {
                const char* proc_name = proc_list[i].kp_proc.p_comm;
                if (proc_name) {
                    if (strcmp(proc_name, processName) == 0) {
                        if (excludeSelf) {
                            if (proc_list[i].kp_proc.p_pid != pid) {
                                ret = proc_list[i].kp_proc.p_pid;
                                break;
                            }
                        } else {
                            ret = proc_list[i].kp_proc.p_pid;
                            break;
                        }
                    }
                }
            }
       
            free(proc_list);
        } 
    }      
    
    return ret;       
}

int checkProcessRunningByPid(pid_t target_pid)
{
    struct kinfo_proc *proc_list = NULL;
    size_t proc_num = 0;
    int ret = 0, i = 0;
    
    if (target_pid) {
        if (getProcessList(&proc_list, &proc_num)) {
            return 0;
            
        } 
        
        for (; i < proc_num; i++) {
            if (proc_list[i].kp_proc.p_pid == target_pid) {
                ret = 1;
                break;
            }
        }
        
        free(proc_list);
    }      
    
    return ret;       
}


int waitForProcessByPid(pid_t target_pid)
{
    struct kinfo_proc *proc_list = NULL;
    size_t proc_num = 0;
    pid_t self = 0; 
    int kq = 0;
	struct kevent ke;
    int i = 0;
    bool process_found = false;
 
    if (getProcessList(&proc_list, &proc_num)) {
        return -1;
    } else {
        self = getpid();

        for (i = 0; i < proc_num; i++) {
            if (proc_list[i].kp_proc.p_pid != self) {
                if (target_pid == proc_list[i].kp_proc.p_pid) {
                    process_found = true;
                    break;
                }
            }
        }    
       
        free(proc_list);
    }
    

    if (!process_found) {
        return 0;
    } else {
        if ((kq = kqueue()) < 0) {
            return -1;
        }

        memset(&ke, 0, sizeof(struct kevent));
        EV_SET(&ke, target_pid, EVFILT_PROC, EV_ADD, NOTE_EXIT, 0, NULL);

	    if (kevent(kq, &ke, 1, NULL, 0, NULL) < 0) {
            return -1;         
        }

        while (true) {
            // this call will block until an event is get for the monitored pid
		    if (kevent(kq, NULL, 0, &ke, 1, NULL) < 0) {
                return -1;
            }

            // process has exited 
            if (ke.fflags & NOTE_EXIT) {
                break;
            }	
        }
    }
    
    return 0;
}

int waitForRunningProcesses(const char* processName)
{
    struct kinfo_proc *proc_list = NULL;
    size_t proc_num = 0;
    pid_t self = 0; 
    pid_t target_pid = 0;
    int kq = 0;
	struct kevent ke;
    int i = 0;
    
    if ((processName == NULL) || (strlen(processName) == 0)) {
        return -1; // EINVAL
    }

    if (getProcessList(&proc_list, &proc_num)) {
        return -1;
    } else {
        struct timeval lastProcStartTime;
        
        memset(&lastProcStartTime, 0, sizeof(struct timeval));
    
        self = getpid();

        for (i = 0; i < proc_num; i++) {
            const char* proc_name = proc_list[i].kp_proc.p_comm;
     
           if (proc_name) {
                if (strcmp(proc_name, processName) == 0) {
                    if (proc_list[i].kp_proc.p_pid != self) {
                        // get the last process started using process 
                        // its start time (avoiding pid numbers wrap up)
                        struct timeval procStartTime = proc_list[i].kp_proc.p_un.__p_starttime;
                        
                        // get the last process started of name "processName"
                        if ((procStartTime.tv_sec > lastProcStartTime.tv_sec) ||
                            ((procStartTime.tv_sec == lastProcStartTime.tv_sec) && 
                             (procStartTime.tv_usec > lastProcStartTime.tv_usec))) {
                            // update timeval struct
                            lastProcStartTime.tv_sec = procStartTime.tv_sec;
                            lastProcStartTime.tv_usec = procStartTime.tv_usec;
                            // set pid to wait for
                            target_pid = proc_list[i].kp_proc.p_pid;
                        } 
                    }
                }
            }
        }    
       
        free(proc_list);
    }
    

    if (target_pid) {
        if ((kq = kqueue()) < 0) {
            return -1;
        }

        memset(&ke, 0, sizeof(struct kevent));
        EV_SET(&ke, target_pid, EVFILT_PROC, EV_ADD, NOTE_EXIT, 0, NULL);

	    if (kevent(kq, &ke, 1, NULL, 0, NULL) < 0) {
            return -1;         
        }

        while (true) {
            // this call will block until an event is get for the monitored pid
		    if (kevent(kq, NULL, 0, &ke, 1, NULL) < 0) {
                return -1;
            }

            // process has exited - continue with monitored pid list
            if (ke.fflags & NOTE_EXIT) {
                break;
            }	
        }
    }
    
    return 0;
}

int killProcess(pid_t target_pid)
{
    int ret = 0;
    
    if (target_pid <= 0) {
        return -1;
    }
    
    ret = kill(target_pid, SIGTERM);
    
    return ret;
}

int getProcessList(struct kinfo_proc **procList, size_t *procCount)
{
    int                 err;
    struct kinfo_proc   *result;
    bool                done;
    static const int    name[] = { CTL_KERN, KERN_PROC, KERN_PROC_ALL, 0 };
    // Declaring name as const requires us to cast it when passing it to
    // sysctl because the prototype doesn't include the const modifier.
    size_t              length;

    *procCount = 0;

    // We start by calling sysctl with result == NULL and length == 0.
    // That will succeed, and set length to the appropriate length.
    // We then allocate a buffer of that size and call sysctl again
    // with that buffer.  If that succeeds, we're done.  If that fails
    // with ENOMEM, we have to throw away our buffer and loop.  Note
    // that the loop causes use to call sysctl with NULL again; this
    // is necessary because the ENOMEM failure case sets length to
    // the amount of data returned, not the amount of data that
    // could have been returned.

    result = NULL;
    done = false;
    do {
        length = 0;
        err = sysctl( (int *) name, (sizeof(name) / sizeof(*name)) - 1,
                      NULL, &length,
                      NULL, 0);
        if (err == -1) {
            err = errno;
        }

        // Allocate an appropriately sized buffer based on the results
        // from the previous call.

        if (err == 0) {
            result = (kinfo_proc*)malloc(length);
            if (result == NULL) {
                err = ENOMEM;
            }
        }

        // Call sysctl again with the new buffer.  If we get an ENOMEM
        // error, toss away our buffer and start again.

        if (err == 0) {
            err = sysctl( (int *) name, (sizeof(name) / sizeof(*name)) - 1,
                          result, &length,
                          NULL, 0);
            if (err == -1) {
                err = errno;
            }
            if (err == 0) {
                done = true;
            } else if (err == ENOMEM) {
				if (result) {
					free(result);
					result = NULL;
				}
                err = 0;
            }
        }
    } while (err == 0 && ! done);

    // Clean up and establish post conditions.

    if (err != 0 && result != NULL) {
        free(result);
        result = NULL;
    }
    *procList = result;
    if (err == 0) {
        *procCount = length / sizeof(struct kinfo_proc);
    }

    return err;
}


int startSyncClientPrefPane()
{
	char* cmdLine[5];
    int res = 0;
	
    cmdLine[0] = (char *)OPEN_COMMAND_PATH;
	cmdLine[1] = (char *)"-a";
    cmdLine[2] = (char *)SYSTEM_PREFERENCES_APP;
    cmdLine[3] = (char *)CLIENT_PREFERENCES_FOLDER;
	cmdLine[4] = NULL;
	
    res = execProcess((char* const*)cmdLine, false);
	
	return res;
}