/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#ifndef INCL_CLIENTCONFIG
#define INCL_CLIENTCONFIG


#include "base/fscapi.h"
#include "base/util/StringBuffer.h"
#include "base/util/StringMap.h"
#include "base/util/ArrayList.h"
#include "client/DMTClientConfig.h"
#include "customization.h"
#include "base/util/PropertyFile.h"

USE_NAMESPACE

/**
 * This class extends DMTClientConfig to store also the default sync path,
 * and to provide the default configuration suited to this client.
 *
 * It implements the Singleton pattern to be available from any point of
 * the application.
 */
class ClientConfig : public DMTClientConfig {
    
private:
    
    /**
     * Map containing the values of the scheduled interval.
     * They are a couple "sourcename", "scheduleIntervalTime"
     */
    StringMap scheduledIntervalValues;
    
    /**
     * The funambol software version. It is used to check the updates
     * of the Funambol product. 
     */
    StringBuffer funambolSwv;

    /**
     * briefcase source path 
     */
    StringBuffer briefCaseFolder;
    
public:
    
    /**
     * Singleton implementation: get the unique instance of the config.
     */
    static ClientConfig *getInstance();
    
    /**
     * Singleton implementation: release the unique instance of the config.
     */
    static void dispose();
    
    // Overloaded methods from DMTClientConfig
    virtual bool read();
    virtual bool save();
    
    // proxy method to have a safe save and not saving during a sync in progress
    bool safeSave();
    
    SyncSourceConfig* getSyncSourceConfig(const char* name);
    SyncSourceConfig* getSyncSourceConfig(int index);
    
    //
    // The following methods of the config read/write inside the config parameters:
    // To save them you must call save() or, only for some of them safeSave is preferred.
    //
    
    /**
     * Stores the value of the schedule interval in a StringMap for the named source
     * To save it can be used the save() or safeSave()
     * @param syncSourceName - the name of the source we want to save the interval
     * @param value - the value (in min) of the schedule interval
     *
     */ 
    void setSchedulingInterval(const char* syncSourceName, const char* value){
        scheduledIntervalValues.put(syncSourceName, value);
    }
    /**
     * get the interval value of the schedule of the named source
     *
     * @param syncSourceName - the name of the source we want to retrieve the interval
     * @return the interval value (in min)
     */
    int getSchedulingInterval(const char* syncSourceName){
        return atoi(scheduledIntervalValues.get(syncSourceName).c_str());
    }
    
    /**
     * set the funambol software version that is the one used to check for the upgrade procedure.
     * It is not used to visualize in the about. For the about check Swv field (in config)
     * To save it must be used save()
     *
     * @param value - the funambol software versione
     */
    void setFunambolSwv(const char* value) {
        funambolSwv = value;
    }
    
    /**
     * @return the funambol software version
     * 
     */
    StringBuffer getFunambolSwv() {
        return funambolSwv;
    }
    
    /**
     * Check if the credential are valid. "Valid" means that 
     * - username or password has a value
     * - the last error code is not 401
     */
    bool areCredentialValid(); 
    
    /**
     * Check if the credential are saved.
     */
    bool areCredentialSaved(); 
    
    //////////////////////////////////////////////////////////////////////////////////////////
    //
    // The following methods of the config write directly inside the config without any
    // other call save() or safeSave()
    //
    
    //
    // Save inside the sync source tree
    //
    void saveLastSyncSourceStatusCode(const char* syncSourceName, int value) {
        StringBuffer s; s.sprintf("%i", value);
        setSyncSourceConfigProperty(syncSourceName, "lastSyncSourceStatusCode", s);
    }
    int readLastSyncSourceStatusCode(const char* syncSourceName){
        const char* s = readSyncSourceConfigProperty(syncSourceName, "lastSyncSourceStatusCode");
        int ret = atoi(s);
        delete [] s;
        return ret;
    }
    
    //
    // save inside the main config
    //
    
    StringBuffer readSyncSourceLastTimestamp(const char* syncSourceName) {
        const char* prop = readSyncSourceConfigProperty(syncSourceName, PROPERTY_SOURCE_LAST_SYNC);
        StringBuffer ret(prop);
        delete [] prop;
        return ret;
    }
    
    /////
    ///// The following methods of the config write inside the main config
    /////
    
    void saveLastSyncStatusCode(int value) {
        StringBuffer s; s.sprintf("%i", value);
        setClientConfigProperty("lastSyncStatusCode", s);
    }
    int readLastSyncStatusCode(){
        const char* s = readClientConfigProperty("lastSyncStatusCode");
        int ret = atoi(s);
        delete [] s;
        return ret;
    }
    
    void saveMenubarUIEnabled(bool value) {
        StringBuffer s = (value ? "true" : "false");
        setClientConfigProperty("menubarUIEnabled", s);
    }
    bool readMenubarUIEnabled() {
        const char* s = readClientConfigProperty("menubarUIEnabled");
        bool ret = false;
        if (s && !strcmp(s, "true")) {
            ret = true;
        }
        delete [] s;
        return ret;
    }
    
    /**
     * Initialize the config: try to read it from file or generate a default one.
     */
    void init();
    
    bool getIsInitialized() {
        return isInitialized;
    }
    
    /**
     * Save only some fields that are in the preference panels
     */
    void savePreferences();
    
    StringBuffer getNewReleaseStringFormat(PropertyFile& pf, bool isFunambol);

    
    void setBriefCaseFolder(const char *newPath);
	const char* getBriefCaseFolder() { return briefCaseFolder; }
    
private:
    
    //------------------------------------------------------------------
    
    /** Singleton implementation: the unique ClientcConfig instance */
    static ClientConfig *instance;
    
    
    void setSyncSourceConfigProperty(const char* syncSourceName, const char* propertyName, const char* propertyValue);
    
    const char* readSyncSourceConfigProperty(const char* syncSourceName, const char* propertyName);
    
    /**
     * get the property that is related to all the client. They are generic and not related to a single source
     * Kind of properties could be push settings, sync error (generic syncml errors...) and so on.
     *
     * @param propertyName - the property name whose value it retrieves
     *
     * @return the value of the propertyName that must be freed
     */
    const char* readClientConfigProperty(const char* propertyName);    
    
    /**
     * set the property that is related to all the client.  They are generic and not related to a single source
     * Kind of properties could be push settings, sync error (generic syncml errors...) and so on.
     *
     * @param propertyName - the property name 
     * @param propertyValue - the property value
     * 
     */
    void setClientConfigProperty(const char* propertyName, const char* propertyValue);

    /**
     * Used to get the release number from the delivered distribution.properties file
     *
     * @param pf - the property file
     * @param isFunambol - true if try ti retrieve the values of funambol_swv property
     *                     false it returns the value of the swv property
     *
     * @return the release number with formula major*10000 + minor * 100 + build
     */
    unsigned long getNewReleaseNumber(PropertyFile& pf, bool isFunambol);
    
    unsigned long getCurrentReleaseNumber(const char* value);
        
    bool isInitialized;
    
    /**
     * Default constructor
     */
    ClientConfig();
    
    /**
     * Generates a default config.
     */
    void createConfig();
    
    /**
     * Save only the account information that are username/password and syncurl.
     * It gets the info from the configuration
     */
    void saveAccountSettings();
    
    /**
     * Save only the info about the log level settings
     */
    void saveLogSettings();

    int setupLog();
};

#endif
