/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <AppKit/AppKit.h>
#import <Cocoa/Cocoa.h>

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <pwd.h>
#include <grp.h>
#include <unistd.h>

#include "utils.h"
#include "constants.h"
#include "customization.h"
#include "base/posixadapter.h"

#import "LaunchCtlConfigManager.h"

#define LAUNCHD_JOBS_CONTROLLER_APP     "launchdJobsController"
#define KICKSTARTER_APP                 "kickstarter"

USE_NAMESPACE

static int installSyncClientSuite(const char* clientInstallDir);
static int startSyncClientJobs(const char* installDir, const char* userName);
static int stopSyncClientJobs(const char* installDir, const char* userName);
static int kickstart(const char* installDir, bool waitForInstaller);
static void closeSystemPrefPanel();
static int checkAndRemoveDir(const char* dirName);
static int chownDir(const char* dirName, uid_t uid, gid_t gid);

int main (int argc, char * const argv[]) 
{
    NSAutoreleasePool* pool = nil;
    const char* clientInstallDir= NULL;
    int c = 0, res = 0;
    int option_index = 0;
    const char* userName    = NULL;
    bool autoDetectAction   = false;
    bool installAction      = false;
    bool upgradeAction      = false;
    bool cleanUpInstallDir  = false;

    struct option long_options[] = {
        {"user",            required_argument, 0, 'u' },
        {"install-dir",     required_argument, 0, 'c' },
        {"detect-action",   no_argument, 0, 'A' },
        {"install",         no_argument, 0, 'I' },
        {"upgrade",         no_argument, 0, 'U' },
        {0, 0, 0, 0}
    };
    
    while ((c = getopt_long(argc, argv, "u:c:AIUC", long_options, &option_index)) != EOF)
    {
        switch (c) {
            case 'u':
                userName = optarg;
                break;
                
            case 'A':
                autoDetectAction = true;
                break;
                
            case 'I':
                installAction = true;
                break;
            
            case 'U':
                upgradeAction = true;
                break;
            
            case 'C':
                cleanUpInstallDir = true;
                break;
                
            case 'c':
                clientInstallDir = optarg;
                break;
                
            case '?':
            default:
                exit(EXIT_FAILURE);
        }
    }
    
    if ((clientInstallDir == NULL) || (strlen(clientInstallDir) == 0)) {
        NSLog(@"no install package directory specified: exiting\n");
        
        exit(EXIT_FAILURE);
    }
    
    if ((userName == NULL) || (strlen(userName) == 0)) {
        NSLog(@"no user name specified: exiting\n");
        
        exit(EXIT_FAILURE);
    }
    
    // detect if sync client is already installed
    // by checking its config
    if (autoDetectAction) {
        StringBuffer configFolder;
        struct stat st;
        
        configFolder.sprintf("/Users/%s/.config/%s/%s", userName, APPNAME, APPCONTEXT);
        
        if (stat(configFolder.c_str(), &st) < 0) {
            if (errno == ENOENT) {
                installAction = true;
                upgradeAction = false;
            } else {
                NSLog(@"error stating \"%s\": %s\n", configFolder.c_str(),
                      strerror(errno));
                
                exit(EXIT_FAILURE);
            }
        } else {
            installAction = false;
            upgradeAction = true;
        }
    }
    
    pool = [[NSAutoreleasePool alloc] init];
    
    // unload all the sync client launchd jobs (menubar and scheduled syncs)
    if (upgradeAction) {
        stopSyncClientJobs(clientInstallDir, userName);
    }
  
    // close System Preferences because we'are going to install a new app for it 
    closeSystemPrefPanel();
    
    // install software package 
    if (installSyncClientSuite(clientInstallDir)) {
        unlink(clientInstallDir);
        
        exit(EXIT_FAILURE);
    }    

    // reload launchd jobs
    if (upgradeAction) {
        startSyncClientJobs(clientInstallDir, userName);
    }
  
    // exec client kickstarter
    if (installAction) {
        if ((res = kickstart(clientInstallDir, true)) != 0) {
            NSLog(@"error executing kickstarter: %s\n", strerror(res)); 
        }
    }
    
    // if specified remove software package installation temporary directory
    if (cleanUpInstallDir) {
        if ((res = removeDir(clientInstallDir)) != 0) {
            NSLog(@"error removing directory \"%s\": %s\n", clientInstallDir, strerror(res));
            return res;
        }
    }
      
    [pool release];

    exit(EXIT_SUCCESS);
}

int installSyncClientSuite(const char* clientInstallDir)
{
    StringBuffer prefPanelTargetPath(CLIENT_PREFERENCES_FOLDER);
    StringBuffer syncEngineTargetPath;
    StringBuffer menuBarTargetPath;
    StringBuffer prefLauncherTargetPath;
    StringBuffer uninstallerTargetPath;
    StringBuffer licenseTargetPath;
    StringBuffer distributionPropTargetPath;
    StringBuffer updaterTargetPath;

    StringBuffer prefPanelSourcePath;
    StringBuffer syncEngineSourcePath;
    StringBuffer menuBarSourcePath;
    StringBuffer prefLauncherSourcePath;
    StringBuffer uninstallerSourcePath;
    StringBuffer licenseSourcePath;
    StringBuffer distributionPropSourcePath;
    StringBuffer updaterSourcePath;
    
    struct group* grp;
    int res = 0;

    if ((clientInstallDir == NULL) || strlen(clientInstallDir) == 0) {
        return EINVAL;
    }

    // format current absolute path for client applications
    prefPanelSourcePath.sprintf("%s/%s",  clientInstallDir, PREFERENCE_APPLICATION_NAME);
    syncEngineSourcePath.sprintf("%s/%s", clientInstallDir, SYNC_APPLICATION_NAME);
    menuBarSourcePath.sprintf("%s/%s.app", clientInstallDir, MENUBAR_APPLICATION_NAME);
    prefLauncherSourcePath.sprintf("%s/%s.app", clientInstallDir, LAUNCHER_APPLICATION_NAME);
    uninstallerSourcePath.sprintf("%s/%s.app", clientInstallDir, UNINSTALLER_APPLICATION_NAME);
    licenseSourcePath.sprintf("%s/%s", clientInstallDir, LICENSE_FILE);
    distributionPropSourcePath.sprintf("%s/%s", clientInstallDir, DISTRIBUITION_PROP_FILE);
    updaterSourcePath.sprintf("%s/%s.app", clientInstallDir, UPDATER_APPLICATION_NAME);
    

    // format final installation names for sync client applications and other stuff
    syncEngineTargetPath.sprintf("%s/%s", CLIENT_SYSTEM_APPLICATION_FOLDER, SYNC_APPLICATION_NAME);
    menuBarTargetPath.sprintf("%s/%s.app", CLIENT_SYSTEM_APPLICATION_FOLDER, MENUBAR_APPLICATION_NAME);
    prefLauncherTargetPath.sprintf("%s/%s.app", CLIENT_USER_APPLICATION_FOLDER, LAUNCHER_APPLICATION_NAME);
    uninstallerTargetPath.sprintf("%s/Uninstall %s.app", CLIENT_USER_APPLICATION_FOLDER, APPNAME);
    licenseTargetPath.sprintf("%s/%s", CLIENT_USER_APPLICATION_FOLDER, LICENSE_FILE);
    distributionPropTargetPath.sprintf("%s/%s", CLIENT_SYSTEM_APPLICATION_FOLDER, DISTRIBUITION_PROP_FILE);
    updaterTargetPath.sprintf("%s/%s.app", CLIENT_SYSTEM_APPLICATION_FOLDER, UPDATER_APPLICATION_NAME);


    if ((res = checkAndRemoveDir(CLIENT_PREFERENCES_FOLDER)) != 0) {
        return res;
    }
    
    if ((res = checkAndRemoveDir(CLIENT_SYSTEM_APPLICATION_FOLDER)) != 0) {
        return res;
    }
    
    if ((res = checkAndRemoveDir(CLIENT_USER_APPLICATION_FOLDER)) != 0) {
        return res;
    }
    
    // create system application directories
    if ((res = mkdir(CLIENT_SYSTEM_APPLICATION_FOLDER, S_IRWXU|S_IRWXG|S_IROTH|S_IXOTH)) < 0) {
        NSLog(@"error creating directory: %s\n", strerror(errno));
        return errno;
    }
    
    if ((res = mkdir(CLIENT_USER_APPLICATION_FOLDER, S_IRWXU|S_IRWXG|S_IROTH|S_IXOTH)) < 0) {
        NSLog(@"error creating directory: %s\n", strerror(errno));
        return errno;
    }

    // install applications moving them to the appropiate system folders 
    // preferences panel
    if ((res = rename(prefPanelSourcePath.c_str(), prefPanelTargetPath.c_str())) < 0) {
        NSLog(@"error renaming \"%s\" to \"%s\": %s\n", prefPanelSourcePath.c_str(), 
              prefPanelTargetPath.c_str(), strerror(errno));
        
        return errno;
    }
    
    // sync engine
    if ((res = rename(syncEngineSourcePath.c_str(), syncEngineTargetPath.c_str())) < 0) {
        NSLog(@"error renaming \"%s\" to \"%s\": %s\n", syncEngineSourcePath.c_str(), 
              syncEngineTargetPath.c_str(), strerror(errno));
        
        return errno;
    }
     
    // menubar application
    if ((res = rename(menuBarSourcePath.c_str(), menuBarTargetPath.c_str())) < 0) {
        NSLog(@"error renaming \"%s\" to \"%s\": %s\n", menuBarSourcePath.c_str(), 
              menuBarTargetPath.c_str(), strerror(errno));
        
        return errno;
    }
    
    // updater application
    if ((res = rename(updaterSourcePath.c_str(), updaterTargetPath.c_str())) < 0) {
        NSLog(@"error renaming \"%s\" to \"%s\": %s\n", updaterSourcePath.c_str(), 
              updaterTargetPath.c_str(), strerror(errno));
        
        return errno;
    }
    
    // preferences panel launcher
    if ((res = rename(prefLauncherSourcePath.c_str(), prefLauncherTargetPath.c_str())) < 0) {
        NSLog(@"error renaming \"%s\" to \"%s\": %s\n", prefLauncherSourcePath.c_str(), 
              prefLauncherTargetPath.c_str(), strerror(errno));
        
        return errno;
    }
    
    // uninstaller
    if ((res = rename(uninstallerSourcePath.c_str(), uninstallerTargetPath.c_str())) < 0) {
        NSLog(@"error renaming \"%s\" to \"%s\": %s\n", uninstallerSourcePath.c_str(), 
              uninstallerTargetPath.c_str(), strerror(errno));
        
        return errno;
    }
    
    // move other files (license, etc)
    if(!POWEREDBY){
        if (rename(licenseSourcePath.c_str(), licenseTargetPath.c_str()) < 0) {
            NSLog(@"error renaming \"%s\" to \"%s\": %s\n", licenseSourcePath.c_str(), 
                licenseTargetPath.c_str(), strerror(errno));
        return errno;  
        }
    }

    if (rename(distributionPropSourcePath.c_str(), distributionPropTargetPath.c_str()) < 0) {
        NSLog(@"error renaming \"%s\" to \"%s\": %s\n", distributionPropSourcePath.c_str(), 
              distributionPropTargetPath.c_str(), strerror(errno));
        
        return errno;
    }

    // get admin group id 
    if ((grp = getgrnam("admin")) == NULL) {
        NSLog(@"error getting admin group info: %s\n", strerror(errno));
        return errno;
    }

    // chown to root:admin
    if ((res = chownDir(CLIENT_PREFERENCES_FOLDER, 0, grp->gr_gid)) != 0) {
        NSLog(@"error changing attributes of directory %s: %s\n", CLIENT_PREFERENCES_FOLDER, strerror(res));
        
        return res;
    }

    if ((res = chownDir(CLIENT_SYSTEM_APPLICATION_FOLDER, 0, grp->gr_gid)) != 0) {
        NSLog(@"error changing attributes of directory %s: %s\n", CLIENT_SYSTEM_APPLICATION_FOLDER, strerror(res));
        
        return res;
    }
 
    if ((res = chownDir(CLIENT_USER_APPLICATION_FOLDER, 0, grp->gr_gid)) != 0) {
        NSLog(@"error changing attributes of directory %s: %s\n", CLIENT_USER_APPLICATION_FOLDER, strerror(res));
        
        return res;
    }  
    
    return res;
}

int kickstart(const char* clientInstallDir, bool waitForInstaller)
{
    int res = 0;
    char* args[3];
    StringBuffer kickstarter;
            
    kickstarter.sprintf("%s/%s", clientInstallDir, KICKSTARTER_APP);
    
    args[0] = (char *)kickstarter.c_str();
    args[1] = (waitForInstaller ? (char *)"-w" : NULL);
    args[2] = NULL;
    
    res = execProcess((char * const *)args, false);
    
    return res;
}

int chownDir(const char* dirName, uid_t uid, gid_t gid)
{
    DIR* dirp = NULL;
    struct dirent *dp = NULL;
    struct stat st;
    int ret = 0;
    
    if ((dirName == NULL) || (strlen(dirName) == 0)) {
        NSLog(@"invalid name\n");
        return EINVAL;
    }
    
    if ((dirp = opendir(dirName)) == NULL) {
        NSLog(@"error opening directory %s: %s\n", dirName, strerror(errno));
        return errno; 
    }
    
    while ((dp = readdir(dirp)) != NULL) {
        if (strcmp(dp->d_name, ".") && strcmp(dp->d_name, "..")) {
            StringBuffer fullPath;
            fullPath.sprintf("%s/%s", dirName, dp->d_name);
            memset(&st, 0, sizeof(struct stat));
            
            if (stat(fullPath.c_str(), &st) < 0) {
                ret = errno;
                NSLog(@"stat - %s: %s\n", fullPath.c_str(), strerror(errno));
                break;
            }
            
            if (S_ISDIR(st.st_mode)) {
                if (chown(fullPath.c_str(), uid, gid) < 0) {
                    NSLog(@"chown - %s: %s\n", fullPath.c_str(), strerror(errno));
                    ret = errno;
                    break;
                }
                
                if ((ret = chownDir(fullPath.c_str(), uid, gid)) != 0) {
                    break;
                }
            } else if (S_ISREG(st.st_mode)) {
                 if (chown(fullPath.c_str(), uid, gid) < 0) {
                     NSLog(@"chown - %s: %s\n", fullPath.c_str(), strerror(errno));
                     ret = errno;
                     break;
                 }
            } else {
                continue;
            }
        }
    }
    
    closedir(dirp);
    
    return ret;
}

int startSyncClientJobs(const char* clientInstallDir, const char *userName)
{
    int res = 0;
    char* args[5];
    StringBuffer launchdJobsController;
    
    if ((userName == NULL) || (strlen(userName) == 0)) {
        NSLog(@"no user name specified: can't stop sync client launchd jobs\n");
        return EINVAL;
    }
    
    launchdJobsController.sprintf("%s/%s", clientInstallDir, LAUNCHD_JOBS_CONTROLLER_APP);
    
    args[0] = (char *)launchdJobsController.c_str();
    args[1] = (char *)"-U";
    args[2] = (char *)userName;
    args[3] = (char *)"-l";
    args[4] = NULL;
    
    res = execProcess((char * const *)args, true);
    
    return res;
}


int stopSyncClientJobs(const char* clientInstallDir, const char *userName)
{
    int res = 0;
    char* args[5];
    StringBuffer launchdJobsController;
    
    if ((userName == NULL) || (strlen(userName) == 0)) {
        NSLog(@"no user name specified: can't stop sync client launchd jobs\n");
        return EINVAL;
    }
    
    launchdJobsController.sprintf("%s/%s", clientInstallDir, LAUNCHD_JOBS_CONTROLLER_APP);
    
    args[0] = (char *)launchdJobsController.c_str();
    args[1] = (char *)"-U";
    args[2] = (char *)userName;
    args[3] = (char *)"-u";
    args[4] = NULL;

    res = execProcess((char *const *)args, true);
        
    return res;
}

int checkAndRemoveDir(const char* dirName)
{
    struct stat st;
    int res = 0;
    
    if ((dirName == NULL) || (strlen(dirName) == 0)) {
        return EINVAL;
    }
        
    memset(&st, 0, sizeof(struct stat));
    
    if (stat(dirName, &st) < 0) {
        if (errno == ENOENT) {
            return res;
        }
        NSLog(@"error stating \"%s\": %s\n", dirName, strerror(errno));
        return errno;
    }
 
    if ((res = removeDir(dirName)) != 0) {
        NSLog(@"error removing directory \"%s\": %s\n", dirName, strerror(errno));
    }
        
    return res;
}

void closeSystemPrefPanel()
{
    NSDictionary* errorDict;
    NSAppleEventDescriptor* returnDescriptor = NULL;
    NSString* sysPrefScript = @"Tell application \"System Preferences\"\nquit\nend tell";
    
    NSAppleScript* scriptObject = [[NSAppleScript alloc] initWithSource: sysPrefScript];
    returnDescriptor = [scriptObject executeAndReturnError: &errorDict];
    [scriptObject release];
}
