/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2008 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */



#include "AddressBookSyncSource.h"

USE_NAMESPACE

void AddressBookSyncSource::error(const char* msg) {
    LOG.error(msg);
}

ABAddressBookRef AddressBookSyncSource::getAddressBook()
{
    if (!addressbook) {
        addressbook = ABGetSharedAddressBook();
        CFRetain(addressbook);
    }
    
    if (!addressbook) {
        error("opening address book");
    }
    
    return addressbook;
}

void AddressBookSyncSource::disposeAddressBook(){
    if(addressbook){
        CFRelease(addressbook);
    }
}

AddressBookSyncSource::AddressBookSyncSource(const char* name, AbstractSyncSourceConfig* sc)
: CacheSyncSource(name, sc, new BlockingSQLiteKeyValueStore(
        getTableName().c_str(), getKeyColumnName().c_str(), getValueColumnName().c_str(),
        getDatabasePath().c_str()
        )
    )
{
    StringBuffer appcontext(APPNAME);
    appcontext += APPCONTEXT;
    PlatformAdapter::init(appcontext);
    addressbook = nil;
    sem_init(&sema,0,1);
    syncsourcelistener = NULL;
}

AddressBookSyncSource::~AddressBookSyncSource()
{
    
    if (saveAddressBook() != 0) {
        LOG.error("Distructor: unable to save AddressBook");
    } else {
        LOG.debug("Distructor: save AddressBook");
    } 
    
    disposeAddressBook();
    sem_destroy(&sema);
}

int AddressBookSyncSource::beginSync() {
    const char* sourcename = config->getName();
    syncsourcelistener = (MacSyncSourceListener*)ManageListener::getInstance().getSyncSourceListener(sourcename);
    return 0;
}

int AddressBookSyncSource::saveAddressBook() {
    
    int ret = 0;
    if (syncsourcelistener && syncsourcelistener->isServerReceivingData()) {
        ABAddressBookRef ab = getAddressBook();
        if (ABHasUnsavedChanges(ab)) {
            if (!ABSave(ab)) {
                LOG.debug("saveAddressBook: addressbook NOT saved");
                ret = -1;
            } else {
                LOG.debug("saveAddressBook: addressbook saved");
            }
        }
    } else {
        LOG.debug("saveAddressBook: not to be saved addressbook");
    }
    return ret;
}


void* AddressBookSyncSource::getItemContent(StringBuffer& key, size_t* size)
{    
    
    ABAddressBookRef ab = getAddressBook();
    CFStringRef uid = CFStringCreateWithCString( kCFAllocatorDefault, key.c_str(), kCFStringEncodingUTF8 );
    ABRecordRef person = ABCopyRecordForUniqueId(ab, uid);
    
    StringBuffer vCard;
    PersonDataConverter *pdc = [[PersonDataConverter alloc] init];
    [pdc personToVcard: (ABPersonRef)person vcard:vCard];
    [pdc release];

    char * ret = stringdup(vCard.c_str());
    *size = strlen(ret);
    CFRelease(person);
    return (void*)ret;
}

Enumeration* AddressBookSyncSource::getAllItemList()
{
    ArrayListEnumeration * en = new ArrayListEnumeration;
    
    ABAddressBookRef ab = getAddressBook();
    CFArrayRef arr = ABCopyArrayOfAllPeople(ab);
    
    int count = CFArrayGetCount(arr);
    
    for (int x = 0; x < count; x++)
    {
        ABPersonRef rec = (ABPersonRef)CFArrayGetValueAtIndex(arr, x);
        CFRetain(rec);
        
        CFStringRef uid = ABRecordCopyUniqueId( rec );
        CFIndex uid_len = CFStringGetLength( uid );
        char *c_uid = new char[ uid_len + 1 ];
        
        CFStringGetCString( uid, c_uid, ( uid_len + 1 ), kCFStringEncodingUTF8 );
        StringBuffer sb(c_uid);
        en->add(sb);
        
        delete [] c_uid;
        
        CFRelease(rec);
        CFRelease(uid);
    }
    
    
    fireSyncSourceEvent(config->getURI(), 
                        config->getName(), 
                        getSyncMode(), count, 
                        SYNC_SOURCE_TOTAL_CLIENT_ITEMS);
        
    
    
    CFRelease(arr);
    return en;
}   

int AddressBookSyncSource::insertItem(SyncItem& item)
{
    CFErrorRef err;
    ABAddressBookRef ab = getAddressBook();
    ABPersonRef abItem = ABPersonCreate();
    
    if (!abItem)
    {
        LOG.error("insertItem: Unable to create person");
        return STC_COMMAND_FAILED;
    }
    
    StringBuffer data((char*)item.getData());
    
    PersonDataConverter *pdc = [[PersonDataConverter alloc] init];
    [pdc vCardToPerson:data person: abItem];
    [pdc release];
    
    lock();
    if (!ABAddRecord(ab, abItem))
    {
        LOG.error("insertItem: Unable to add Person to AddressBook for syncItem with key %s", item.getKey());
        CFRelease(abItem);
        CFRelease(err);
        unlock();
        return STC_COMMAND_FAILED;
    }
    /*
    if (!ABSave(ab))
    {
        LOG.error("insertItem: Unable to save AddressBook for syncItem with key %s", item.getKey());
        CFRelease(abItem);
        CFRelease(err);
        unlock();
        return STC_COMMAND_FAILED;
    }
    */
    CFStringRef uid = ABRecordCopyUniqueId(abItem);
    CFIndex uid_len = CFStringGetLength( uid );
    char *c_uid = new char[ uid_len + 1 ];
    CFStringGetCString( uid, c_uid, ( uid_len + 1 ), kCFStringEncodingUTF8 );
    
    item.setKey(c_uid);
    
    CFRelease(abItem);
    CFRelease( uid );
    delete [] c_uid;
    
    LOG.info("insertItem: Success for item with key %s", item.getKey());
    unlock();
    
    return STC_ITEM_ADDED;
}   


int AddressBookSyncSource::modifyItem(SyncItem& item)
{
    CFErrorRef err;
    
    lock();
    
    ABAddressBookRef ab = getAddressBook();
    CFStringRef itemKey = Std2CFString(item.getKey());
    ABRecordRef abItem  = ABCopyRecordForUniqueId(ab, itemKey);
    CFRetain(abItem);

    StringBuffer temp((char*)item.getData());

    PersonDataConverter *pdc = [[PersonDataConverter alloc] init];
    [pdc vCardToPerson:temp person:(ABPersonRef)abItem];
    [pdc release];
    
    if (!abItem)
    {
        LOG.error("modifyItem: Cannot get item with key %s", item.getKey());
        return STC_COMMAND_FAILED;
    }
    
    lock();
    
    /*
    if (!ABSave(ab))
    {
        LOG.error("modifyItem: Unable to save Person or AddressBook for syncItem with key %s", item.getKey());
        // TODO : Handle error
        CFRelease(abItem);
        CFRelease(err);
        unlock();
        return STC_COMMAND_FAILED;
    }
     */
    CFRelease(abItem);
    LOG.info("modifyItem: Success for item with key %s", item.getKey());
    
    unlock();
    
    return STC_OK;
}   

int AddressBookSyncSource::removeItem(SyncItem& item)
{
    CFErrorRef err;
    
    ABAddressBookRef ab = getAddressBook();
    const char* itemKeyStr = item.getKey();
    CFStringRef itemKey = Std2CFString(itemKeyStr);
    ABRecordRef abItem  = ABCopyRecordForUniqueId(ab, itemKey);
        
    if (!abItem) {
        LOG.info("removeItem: contact not found (id %s): already deleted by user?", itemKeyStr);
        return STC_OK;
    }

    CFRetain(abItem);
    lock();
    
    if (!ABRemoveRecord(ab, abItem)) {
        LOG.error("removeItem: Unable to remove Person or save AddressBook for syncItem with key %s", itemKeyStr);
        CFRelease(err);
        CFRelease(abItem);
        unlock();

        return STC_ITEM_NOT_DELETED;
    }

    LOG.debug("removeItem: successfully removed item with key %s", itemKeyStr);
    CFRelease(abItem);
    unlock();

    return STC_OK;
}      
/*
StringBuffer AddressBookSyncSource::getItemSignature(StringBuffer& key) {

    ABAddressBookRef ab = getAddressBook();
    CFStringRef itemKey = Std2CFString(key);
    ABRecordRef abItem  = ABCopyRecordForUniqueId(ab, itemKey);
    
    if (abItem)
    {
        CFRetain(abItem);
        StringBuffer ret = getLastModificationTime(abItem);
        CFRelease(abItem);
        return ret;
    }
    else
    {
        StringBuffer ret = "-1"; // Deleted?
        return ret;
       
    }
    
}
*/
int AddressBookSyncSource::removeAllItems()
{
    lock();

    CFErrorRef err;
    
    ABAddressBookRef ab = getAddressBook();
    CFArrayRef arr = ABCopyArrayOfAllPeople(ab);
    
    int count = CFArrayGetCount(arr);
    
    for (int x = 0; x < count; x++)
    {
        ABPersonRef uid = (ABPersonRef)CFArrayGetValueAtIndex( arr, x );
        CFRetain(uid);        
        if (!ABRemoveRecord(ab, (ABRecordRef)uid))
        {
            // TODO : Handle error
            LOG.error("removeAllItems: error removing record");
            CFRelease(err);
            CFRelease(uid);
            unlock();
            
            return 1;
        }
        
        CFRelease(uid);
    }
    if (!ABSave(ab))
    {
        LOG.error("removeAllItem: Unable to save all deleted items");
        CFRelease(err);
        unlock();
        return 1;
    }
    
    clearCache();
    
    setLastAnchor("");
    setNextAnchor("");
    config->setLast(1);

    unlock();
    return 0;
}

void AddressBookSyncSource::stopSync()
{
    endSync(false);
}

/*
const StringBuffer AddressBookSyncSource::getLastModificationTime(ABRecordRef person)
{
    if (!person)
    {
        error("Cannot get last modified time of null person");
        return StringBuffer("0");
    }
    
    PersonDataConverter *pdc = [[PersonDataConverter alloc] init];
    CFStringRef str = [pdc getPersonProperty: (ABPersonRef)person property: kABModificationDateProperty];
    [pdc release];
    
    StringBuffer ret = CFString2Std(str);
    CFRelease(str);
    return ret;

}
*/

int AddressBookSyncSource::endSync(bool shouldUnlock)
{
    lock();
    
    saveCache();
    
    if (!shouldUnlock)
    {
        LOG.info("Will not unlock semaphore in endSync()");
        return 0;
    }

    unlock();
    
    // Wait for 1 second to make sure modification timestamps are before the anchor
    sleep(1000);
    
    return 0;
}




// Static

int AddressBookSyncSource::initializeDatabase()
{
    sqlite3 * db;
    int ret;
    
    ret = sqlite3_open(getDatabasePath().c_str(), &db);
    if (ret != SQLITE_OK)
    {
        setErrorF(ret, "Unable to open SQLite database: %s", getDatabasePath().c_str());
    }

    sqlite3_stmt * stmt;
    
    StringBuffer sql("");
    sql.sprintf("CREATE TABLE %s (%s TEXT PRIMARY KEY, %s TEXT)", getTableName().c_str(), getKeyColumnName().c_str(), getValueColumnName().c_str());
    
    ret = sqlite3_prepare(db, sql.c_str(), sql.length(), &stmt, NULL);
    if (ret != SQLITE_OK) {
        setErrorF(ret, "Unable to prepare create SQLite table in %s: %s", getDatabasePath().c_str(), getTableName().c_str());
        return ret;
    }
    ret = sqlite3_step(stmt);
    if (ret != SQLITE_OK && ret != SQLITE_DONE) {
        setErrorF(ret, "Unable to execute create SQLite table in %s: %s", getDatabasePath().c_str(), getTableName().c_str());
        return ret;
    }
    sqlite3_finalize(stmt);
    
    sqlite3_close(db);

    return 0;
}

/**
 * Verify the SQLite database used by the cache exists
 *
 * @return Zero on success, non-zero on failure
 **/
int AddressBookSyncSource::verifyDatabase()
{
    sqlite3 * db;
    int ret;
    ret = sqlite3_open(getDatabasePath().c_str(), &db);
    if (ret != SQLITE_OK)
    {
        setErrorF(ret, "Unable to open SQLite database: %s", getDatabasePath().c_str());
        return ret;
    }
    
    char ** temp;
    int rows;
    int cols;
    
    StringBuffer sql("");
    sql.sprintf("SELECT %s, %s FROM %s LIMIT 0", getKeyColumnName().c_str(), getValueColumnName().c_str(), getTableName().c_str());
    LOG.debug("Verify Database query: %s", sql.c_str());
    ret = sqlite3_get_table(db, sql.c_str(), &temp, &rows, &cols, NULL);
    if (ret != SQLITE_OK)
    {
        setErrorF(ret, "Unable to verify SQLite table in %s: %s", getDatabasePath().c_str(), getTableName().c_str());
    }
    sqlite3_close(db);
    
    
    return ret;
}


StringBuffer AddressBookSyncSource::getDatabasePath()
{
    StringBuffer dataBasePath = PlatformAdapter::getConfigFolder(); 
    dataBasePath.append(SQLITE_DB_NAME);
    LOG.debug("AddressBookSyncSource database setted in: %s", dataBasePath.c_str());
   
    return dataBasePath;
}

StringBuffer AddressBookSyncSource::getTableName()
{
    StringBuffer sb("addressbook");
    return sb;
}

StringBuffer AddressBookSyncSource::getValueColumnName()
{
    StringBuffer sb("lastModified");
    return sb;
}

StringBuffer AddressBookSyncSource::getKeyColumnName()
{
    StringBuffer sb("personID");
    return sb;
}


void AddressBookSyncSource::lock()
{
    sem_wait(&sema);
}

void AddressBookSyncSource::unlock()
{
    sem_post(&sema);
}
