/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "SyncProcessManager.h"
#include "constants.h" // costants and defines for Funambol sync clien and UI components

USE_NAMESPACE

#define WAITING_TIME_BEFORE_KILL_SYNC_PROCESS   5 // in seconds

/**
 * callback function that receives the notification from the Menubar and call the cancelSync method
 * of the class.
 */ 
void cancelSyncFromMenubarCallBack(CFNotificationCenterRef center, void *observer, 
                                   CFStringRef name, const void *object, CFDictionaryRef userInfo);

/**
 * exit the sync process if the soft cancel is not working properly
 */
void exitSyncProcess(int syncStatus);

void killSyncProcess();

@implementation WaitQueueHandler

- (id) init
{
    self = [super init];
    return self;    
}

- (void) setQueueManager: (SyncQueueManager *) queueManager
{
    waitQueueManager = queueManager;
}

- (void) resetQueue
{
    waitQueueManager->resetSyncQueue();
}   
@end


/**
 * signal handler thread: trap signals and perform a cancel sync
 */
@implementation SigHandler

- (id) init
{
    if (self = [super init]) {
        threadRunning = NO;
    }
    
    return self;
}

- (void) runSyncSigHandler: (SyncProcessManager *) syncManager
{
    int i = 0, signals_num = 0, kq = 0;
    // a list of signals (from kill -l) to trap. For now they are
    // catched and a cancel sync action is performed: maybe some of
    // them (such as SIGUSR1 and SIGUSR2) could be reserved in
    // future for special beahviour of the client
    int signals[] = { SIGHUP,  SIGINT,  SIGQUIT, SIGILL,  SIGTRAP, 
                      SIGABRT, SIGALRM, SIGBUS,  SIGSEGV, SIGPIPE, 
                      SIGTERM, SIGCHLD, SIGUSR1, SIGUSR2 };
    
    NSAutoreleasePool *pool = nil;
	
    signals_num = sizeof(signals) / sizeof(int);
    
    if ((kq = kqueue ()) < 0) {
        LOG.error("error creating kqueue: %s", strerror(errno));
        return;
    }
    
    // register each signal event with the kqueue
    for (; i < signals_num; i++)  {
        struct kevent event;
        EV_SET (&event, signals[i], EVFILT_SIGNAL, EV_ADD | EV_ENABLE, 0, 0, NULL);
        
        // ignore signal
        signal(signals[i], SIG_IGN);
        
        // register the signal event
        if (kevent(kq, &event, 1, NULL, 0, NULL) == -1) {
            LOG.error("could not register kevent: %s\n", strerror(errno));
            return;
        }
    }
   
    pool = [[NSAutoreleasePool alloc] init];
    
    threadRunLoop = CFRunLoopGetCurrent();
    
    CFRunLoopRun();
    threadRunning = YES;
    
    while (1) {
        int status;
        struct kevent event;
        
        memset(&event, 0, sizeof(struct kevent));
               
        if ((status = kevent (kq, NULL, 0, &event, 1, NULL)) < 0) {
            LOG.error("error waiting for kevent: %s", strerror(errno));
            break;
            
        }
        
        if (status > 0) {
            // signal catched
            signalCatched = YES;
            LOG.info("signal catched (%s)", strsignal((int)event.ident));
            // try cancelling sync
            [syncManager cancelSync];
        }
    }
    
    CFRunLoopStop(threadRunLoop);
    threadRunning = NO;
    
    [pool release];
}

- (void) runWaitQueueSigHandler: (WaitQueueHandler *)queueManager
{
    int i = 0, signals_num = 0, kq = 0;
    // a list of signals (from kill -l) to trap. For now they are
    // catched and a cancel sync action is performed: maybe some of
    // them (such as SIGUSR1 and SIGUSR2) could be reserved in
    // future for special beahviour of the client
    int signals[] = { SIGHUP,  SIGINT,  SIGQUIT, SIGILL,  SIGTRAP, 
        SIGABRT, SIGALRM, SIGBUS,  SIGSEGV, SIGPIPE, 
        SIGTERM, SIGCHLD, SIGUSR1, SIGUSR2 };
    
    NSAutoreleasePool *pool = nil;
    
    signals_num = sizeof(signals) / sizeof(int);
    
    signalCatched = NO;
    
    if ((kq = kqueue ()) < 0) {
        LOG.error("error creating kqueue: %s", strerror(errno));
        return;
    }
    
    // register each signal event with the kqueue
    for (; i < signals_num; i++)  {
        struct kevent event;
        EV_SET (&event, signals[i], EVFILT_SIGNAL, EV_ADD | EV_ENABLE, 0, 0, NULL);
        
        // ignore signal
        signal(signals[i], SIG_IGN);
        
        // register the signal event
        if (kevent(kq, &event, 1, NULL, 0, NULL) == -1) {
            LOG.error("could not register kevent: %s\n", strerror(errno));
            return;
        }
    }
    
    pool = [[NSAutoreleasePool alloc] init];
    
    threadRunLoop = CFRunLoopGetCurrent();
    
    CFRunLoopRun();
    threadRunning = YES;
    
    while (1) {
        int status;
        struct kevent event;
        
        memset(&event, 0, sizeof(struct kevent));
        
        if ((status = kevent (kq, NULL, 0, &event, 1, NULL)) < 0) {
            LOG.error("error waiting for kevent: %s", strerror(errno));
            break;
            
        }
        
        if (status > 0) {
            // signal catched
            signalCatched = YES;
            LOG.info("signal catched (%s)", strsignal((int)event.ident));
            [queueManager resetQueue];
            
            break;
        }
    }
    
    CFRunLoopStop(threadRunLoop);
    threadRunning = NO;
    
    [pool release];
}


- (void) stopSigHandler
{   
    if (threadRunning) {
        CFRunLoopStop(threadRunLoop);
    }
}

- (BOOL) getSignalCatched
{
    return signalCatched;
}

@end

@implementation SyncProcessManager

- (id) init
{
    // register the class to receive the notification from the manubar
    if (self = [super init]) {
        nCenter = CFNotificationCenterGetDistributedCenter();
        CFNotificationCenterAddObserver(nCenter, 
                                        (id)self, 
                                        cancelSyncFromMenubarCallBack,     
                                        CFSTR(SYNC_CANCELLED_NOTIFICATION),
                                        NULL,                                     
                                        CFNotificationSuspensionBehaviorDeliverImmediately);
        syncStatus = 0;
    }
    
    return (self);
}

- (void) dealloc
{
    CFNotificationCenterRemoveObserver(nCenter, (id)self, CFSTR(SYNC_CANCELLED_NOTIFICATION), NULL);
    
    [super dealloc];
} 

- (void) runSyncThread: (NSMutableArray *) sourcesNameList
{
    if ([sourcesNameList count]) {
        NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    
		[self startSyncProcessNotification];
        
		for (NSString *sourceName in sourcesNameList) {
            if ([sourceName length]) {
                syncStatus = [self doSync:[sourceName UTF8String]];
            }
        }
        
        [pool drain];
        
        exitSyncProcess(syncStatus);
    }
} 

- (void) runResetThread: (NSMutableArray *) sourcesNameList
{
    if ([sourcesNameList count]) {
        NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
        ClientConfig* config = ClientConfig::getInstance();
 
		[self startSyncProcessNotification];

        for (NSString *sourceName in sourcesNameList) {
            const char* sName = [sourceName UTF8String];
            
            if (sName) {
                SyncSourceConfig *srcConfig = config->getSyncSourceConfig(sName);
                
                if (srcConfig) {
                    AddressBookSyncSource abss(sName, srcConfig);
               
                    // reset source anchor
                    srcConfig->setLast(1); 
                    // save anchor before cleanup start: this
                    // will make sure that next sync will be a slow 
                    // even if something goes wrong
                    config->save();        
                    abss.removeAllItems();
			
                    syncStatus = [self doSync:sName];
                }
            }
        }
        
        [pool drain];
        
        exitSyncProcess(syncStatus);
    }
} 

-(void) cancelSync
{
	ClientConfig* config = ClientConfig::getInstance();

	NSLog(@"Notification received by cancelSync");
	config->setToAbort();
	
    // now wait for some times and then kill this thread
    CFAllocatorRef allocator = kCFAllocatorDefault;
    CFAbsoluteTime fireDate = CFAbsoluteTimeGetCurrent() + WAITING_TIME_BEFORE_KILL_SYNC_PROCESS;
    CFTimeInterval interval = 0;
    CFOptionFlags flags = 0;
    CFIndex order = 0;
        
    CFRunLoopTimerCallBack callback = (CFRunLoopTimerCallBack)killSyncProcess;
    CFRunLoopTimerRef timer = CFRunLoopTimerCreate(allocator, fireDate, interval, flags, order, callback, NULL);
    CFRunLoopAddTimer(CFRunLoopGetCurrent(), timer, kCFRunLoopDefaultMode);
}


- (int) doSync: (const char*) sourcename
{
    ClientConfig* config = ClientConfig::getInstance();
    SyncSourceConfig *srcconfig = config->getSyncSourceConfig(sourcename);
    SyncSource* ss;
	// Create the SyncClient
    SyncClient syncClient;
	
    MacSyncItemListener* itemListener = new MacSyncItemListener(sourcename);
    MacSyncSourceListener* sourceListener = new MacSyncSourceListener(sourcename);
    MacSyncListener* syncListener = new MacSyncListener(sourcename);
    MacTransportListener* transportListener = new MacTransportListener(sourcename);
    
    ManageListener::getInstance().setSyncItemListener(itemListener);
    ManageListener::getInstance().setSyncSourceListener(sourceListener);
    ManageListener::getInstance().setSyncListener(syncListener);
    ManageListener::getInstance().setTransportListener(transportListener);
    
    // Create the SyncSource passing its name, the SyncSourceConfig 
    if (strcmp(sourcename, CONTACTS_SOURCE_NAME) == 0 ) {
        ss = new AddressBookSyncSource(sourcename, srcconfig);        
    } else if(strcmp(sourcename, CONFIG_SOURCE_NAME) == 0 ) {
        ss = new ConfigSyncSource(sourcename, PlatformAdapter::getConfigFolder(), srcconfig);
	} else if(strcmp(sourcename, BRIEFCASE_SOURCE_NAME) == 0 ) {
		ss = new FileSyncSource(sourcename, srcconfig);
		((FileSyncSource *)ss)->setDir(config->getBriefCaseFolder());
	}
    
	if (ss) {
		transportListener->setSyncSourceReference(ss);
	} else {
		LOG.error("missing sync source");
		return kSourceDefinitionNotFound;
	}
	    
    // Initialize the SyncSource array to sync
    SyncSource* ssArray[] = { ss, NULL } ;
    
    //SyncSourceConfig* ssconfig = config->getSyncSourceConfig(sourcename);
	// get the last time stamp for this source
    int last = srcconfig->getLast();
    
    // post sync source name notification
	[self postSourceNameNotification: sourcename];

    // SYNC!
    if ((syncStatus = syncClient.sync(*ClientConfig::getInstance(), ssArray)) != 0) {
        LOG.error("error during sync [code %d]", syncStatus);
        // keep the old value of the sync
        srcconfig->setLast(last);
        // @TODO: try to insert the data of the currenty sync source inside the config 
        // of that source using methods like getLastSyncSourceStatusCode
        // Actually it is possible to add other info about the syncReport
    } 
    
	// set the last sync status code of the whole sync process
    config->saveLastSyncStatusCode(syncStatus);
    config->save();
    
    SyncReport* sreport = syncClient.getSyncReport();
    if (sreport) {
        StringBuffer rep;
        sreport->toString(rep, true);
        LOG.info("\n%s", rep.c_str());
    }
    
    ManageListener::getInstance().unsetSyncItemListener(sourcename);
    ManageListener::getInstance().unsetSyncSourceListener(sourcename);
    ManageListener::getInstance().unsetSyncListener(sourcename);
    ManageListener::getInstance().unsetTransportListener(sourcename);
    
    delete ss;

    return syncStatus;
}

- (void) postSourceNameNotification: (const char*) sourcename
{
    CFStringRef  sourceName = CFStringCreateWithCString(NULL, sourcename, kCFStringEncodingUTF8);
    CFMutableDictionaryRef infoDict = CFDictionaryCreateMutable(NULL, 1, &kCFCopyStringDictionaryKeyCallBacks, 
                                                                &kCFTypeDictionaryValueCallBacks);
    CFDictionaryAddValue(infoDict, CFSTR("SourceName"), sourceName);
    CFNotificationCenterPostNotification(nCenter, CFSTR(SYNC_SOURCE_NAME_NOTIFICATION), NULL, infoDict, TRUE);
    
    CFRelease(sourceName);
    CFRelease(infoDict);
}

-(void) startSyncProcessNotification
{
    NSLog(@"posting %@ notification", CFSTR(CLIENT_SYNC_START_NOTIFICATION));
    pid_t pid = getpid();
    CFNumberRef syncProcessPid = CFNumberCreate(NULL, kCFNumberIntType, (int *)&pid);
    CFMutableDictionaryRef infoDict = CFDictionaryCreateMutable(NULL, 1, &kCFCopyStringDictionaryKeyCallBacks, 
                                                                &kCFTypeDictionaryValueCallBacks);
    
    CFDictionaryAddValue(infoDict, CFSTR("SyncProcessPid"), syncProcessPid);
    
    CFNotificationCenterPostNotification(nCenter, CFSTR(CLIENT_SYNC_START_NOTIFICATION), NULL, infoDict, YES);
    
    CFRelease(syncProcessPid);
    CFRelease(infoDict);
}

@synthesize syncStatus;

@end 


// called only by the timer when killing the sync process
void killSyncProcess()
{
    ClientConfig* config = ClientConfig::getInstance();
							
    config->saveLastSyncStatusCode(kSyncCancelledByUser);
    config->save();
							
    exitSyncProcess(kSyncCancelledByUser);
}


//
// Functions to exit the sync process (by exiting the main loop) and that receiving the 
// command from the menubar
//

void exitSyncProcess(int syncStatus)
{
    CFNotificationCenterRef nCenter = CFNotificationCenterGetDistributedCenter();
    CFMutableDictionaryRef infoDict = CFDictionaryCreateMutable(NULL, 1, &kCFCopyStringDictionaryKeyCallBacks, 
                                                                &kCFTypeDictionaryValueCallBacks);
    CFNumberRef syncStatusCode      = CFNumberCreate(NULL, kCFNumberIntType, &syncStatus);
    
    CFDictionaryAddValue(infoDict, CFSTR("SyncStatusCode"), syncStatusCode);
    
    CFNotificationCenterPostNotification(nCenter, CFSTR(CLIENT_SYNC_ENDED_NOTIFICATION), NULL, infoDict, YES);
    CFRelease(syncStatusCode);
    CFRelease(infoDict);
    
    CFRunLoopRef r = CFRunLoopGetMain();
    CFRunLoopStop(r);
}


void cancelSyncFromMenubarCallBack(CFNotificationCenterRef center, void *observer, 
                              CFStringRef name, const void *object, CFDictionaryRef userInfo)
{
    NSLog(@"Notification received by cancelSyncFromMenubarCallBack");
    [(id)observer cancelSync];
} 

