#!/bin/bash

log_level_error=0
log_level_info=1
log_level_debug=2

function check_sync_log()
{
    if [ -f $sync_log ]; then
        return 1
    fi

    return 0
}

function check_sync_log_debug_level()
{
    check_sync_log_pattern "Beginning synchronization"

    return $?
}

function check_sync_log_info_level()
{
    check_sync_log_pattern "Preparing synchronization of addressbook..."

    if [ $? != 0 ]; then
        return $EXIT_FAILURE
    fi

    check_sync_log_pattern "Beginning synchronization"

    if [ $? == 0 ]; then
        return $EXIT_FAILURE
    fi

    return $EXIT_SUCCESS
}

function check_sync_log_error_level()
{
    check_sync_log_pattern "Preparing synchronization of addressbook..."

    if [ $? == 0 ]; then
        return $EXIT_FAILURE
    fi

    check_sync_log_pattern "Beginning synchronization"

    if [ $? == 0 ]; then
        return $EXIT_FAILURE
    fi

    return $EXIT_SUCCESS
}

function check_sync_log_pattern()
{
    pattern_string="$1"

    if [ -z "$pattern_string" ]; then
        return $EXIT_FAILURE
    fi 

    if [ ! -f $sync_log ]; then
        return $EXIT_FAILURE
    fi

    # check if a predefined debug string is in log:
    # grep will return 0 if string is present, 1 if
    # it isn't present in logfile, 2 for other errors
    grep "$pattern_string" $sync_log >/dev/null 2>&1 

    return $?
}

function check_debug_log()
{
    # before starting test, try to stop all sync process running 
    cleanup_sync_running
    
    # check and remove pre-existing sync client log files
    check_sync_log

    if [ $? == 1 ]; then
        rm $sync_log

        if [ $? != 0 ]; then
            log_error "can't remove sync client log file" 

            return $EXIT_FAILURE   
        fi
    fi

    # set sync client log level to debug and then start a sync:
    # at sync client exit we must find a new log file written
    set_log_level $log_level_debug

    if [ $? != 0 ]; then
        log_error "error setting sync client log to debug level"

        return $EXIT_FAILURE
    fi

    sync_proc_pid=$(sync_contacts 0)

    if [ -z "$sync_proc_pid" ]; then
        log_error "error staring sync process"

        return $EXIT_FAILURE
    fi

    wait_sync_start
    cleanup_sync_running
    check_sync_log_debug_level

    if [ $? != 0 ]; then
        log_error "missing DEBUG level statements in log"
        rm $sync_log
        
        return $EXIT_FAILURE
    fi

    # cleanup log after test
    rm $sync_log

    return $EXIT_SUCCESS
}

function check_info_log()
{
    # before starting test, try to stop all sync process running 
    cleanup_sync_running
    
    # check and remove pre-existing sync client log files
    check_sync_log

    if [ $? == 1 ]; then
        rm $sync_log

        if [ $? != 0 ]; then
            log_error "can't remove sync client log file" 

            return $EXIT_FAILURE   
        fi
    fi

    # set sync client log level to debug and then start a sync:
    # at sync client exit we must find a new log file written
    set_log_level $log_level_info

    if [ $? != 0 ]; then
        log_error "error setting sync client log to debug level"

        return $EXIT_FAILURE
    fi

    sync_proc_pid=$(sync_contacts 0)

    if [ -z "$sync_proc_pid" ]; then
        log_error "error staring sync process"

        return $EXIT_FAILURE
    fi

    wait_sync_start
    cleanup_sync_running
    check_sync_log_info_level

    if [ $? != 0 ]; then
        log_error "log level seem not to be set to info"
        
        return $EXIT_FAILURE
    fi

    # cleanup log after test
    rm $sync_log

    return $EXIT_SUCCESS
}

function check_error_log()
{
    # before starting test, try to stop all sync process running 
    cleanup_sync_running
    
    # check and remove pre-existing sync client log files
    check_sync_log

    if [ $? == 1 ]; then
        rm $sync_log

        if [ $? != 0 ]; then
            log_error "can't remove sync client log file" 

            return $EXIT_FAILURE   
        fi
    fi

    # set sync client log level to debug and then start a sync:
    # at sync client exit we must find a new log file written
    set_log_level $log_level_error

    if [ $? != 0 ]; then
        log_error "error setting sync client log to debug level"

        return $EXIT_FAILURE
    fi

    sync_proc_pid=$(sync_contacts 0)

    if [ -z "$sync_proc_pid" ]; then
        log_error "error staring sync process"

        return $EXIT_FAILURE
    fi

    wait_sync_start
    cleanup_sync_running
    check_sync_log_error_level

    if [ $? != 0 ]; then
        log_error "log level seem not to be set to ERROR"
        rm $sync_log
        
        return $EXIT_FAILURE
    fi

    # cleanup log after test
    rm $sync_log

    return $EXIT_SUCCESS
}

function check_log_config()
{
    # before starting test, try to stop all sync process running 
    cleanup_sync_running
    
    # check and remove pre-existing sync client log files
    check_sync_log

    if [ $? == 1 ]; then
        rm $sync_log

        if [ $? != 0 ]; then
            log_error "can't remove sync client log file" 

            return $EXIT_FAILURE   
        fi
    fi

    # set sync client log level to debug and then start a sync:
    # at sync client exit we must find a new log file written
    set_log_level $log_level_debug

    if [ $? != 0 ]; then
        log_error "error setting sync client log to debug level"

        return $EXIT_FAILURE
    fi

    sync_proc_pid=$(sync_contacts 0)

    if [ -z "$sync_proc_pid" ]; then
        log_error "error staring sync process"
        rm -f $sync_log

        return $EXIT_FAILURE
    fi

    wait_sync_start
    cleanup_sync_running
    check_sync_log

    if [ $? == 0 ]; then
        log_error "can't find sync client log file" 
        
        rm -f $sync_log

        return $EXIT_FAILURE
    fi

    # cleanup log after test
    rm $sync_log

    # stop sync process
    cleanup_sync_running

    return $EXIT_SUCCESS
}

function main() 
{ 
    # source in unit tests common utilities
    source unit-test-common

    if [ $? != 0 ]; then
        echo >&2 "can't source in unit-test-common: exiting..."

        exit 1
    fi 

    log_message "starting sync log unit test..."

    if [ ! -x "$sync_engine" ]; then
        log_error "sync engine application not found: check build output in $build_dir or consider 
            exporting FUNAMBOL_SYNC_ENGINE in your enviroment"
    
        exit $EXIT_FAILURE
    fi
 
    ret=$EXIT_SUCCESS 
    ## check if sync log file is written in the correct place
    check_log_config
    if [ $? != 0 ]; then
        echo >&2 "check_log test failed"
        ret=$EXIT_FAILURE
    fi 

    check_error_log
    if [ $? != 0 ]; then
        echo >&2 "check_log test failed"
        ret=$EXIT_FAILURE
    fi 

    check_info_log
    if [ $? != 0 ]; then
        echo >&2 "check_log test failed"
        ret=$EXIT_FAILURE
    fi 

    check_debug_log
    if [ $? != 0 ]; then
        echo >&2 "check_log test failed"
        ret=$EXIT_FAILURE
    fi 

    exit $ret
}

main $@
# EOF
